/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * Portions Copyrighted 2009 Jaroslav Tulach
 */

package cz.xelfi.quoridor.webidor.resources;

import cz.xelfi.quoridor.IllegalPositionException;
import cz.xelfi.quoridor.webidor.*;
import cz.xelfi.quoridor.Move;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.ws.rs.DefaultValue;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.PUT;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response.Status;

/**
 *
 * @author Jaroslav Tulach <jtulach@netbeans.org>
 */
public final class Games {
    private final Quoridor quoridor;
    private final List<Game> games = new ArrayList<Game>();
    private final File dir;
    private static final Logger LOG = Logger.getLogger(Games.class.getName());

    public Games(File dir, Quoridor quoridor) {
        this.dir = dir;
        this.quoridor = quoridor;
        File[] arr = dir.listFiles();
        if (arr != null) {
            for (File f : arr) {
                try {
                    Game g = readGame(f);
                    games.add(g);
                } catch (IOException ex) {
                    LOG.log(Level.WARNING, "Wrong game in " + f, ex);
                }
            }
        }
    }

    @POST
    @Produces({ MediaType.APPLICATION_JSON, MediaType.TEXT_XML })
    public GameId createGame(
        @QueryParam("loginID") String id,
        @QueryParam("white") String white,
        @QueryParam("black") String black
    ) throws IOException {
        String logUser = quoridor.isLoggedIn(id);
        if (logUser == null) {
            throw new WebApplicationException(Status.UNAUTHORIZED);
        }
        if (white == null) {
            throw new WebApplicationException(Status.PRECONDITION_FAILED);
        }
        if (black == null) {
            throw new WebApplicationException(Status.PRECONDITION_FAILED);
        }
        if (!logUser.equals(white) && !logUser.equals(black)) {
            throw new WebApplicationException(Status.PRECONDITION_FAILED);
        }

        Game g = new Game(white, black);
        storeGame(g);
        games.add(g);
        return g.getId();
    }

    @GET
    @Path("{id}")
    @Produces(MediaType.TEXT_PLAIN)
    public String getBoard(
        @PathParam("id") String id,
        @QueryParam("move") @DefaultValue("-1") int move
    ) {
        Game g = findGame(id, move);
        if (g == null) {
            throw new IllegalArgumentException("Unknown game " + id);
        }
        return g.getBoard().toString();
    }

    @GET
    @Path("{id}")
    @Produces({ MediaType.APPLICATION_JSON, MediaType.TEXT_XML })
    public Game getBoardInfo(
        @QueryParam("loginID") @DefaultValue("") String loginId,
        @PathParam("id") String id,
        @QueryParam("move") @DefaultValue("-1") int move
    ) throws IOException {
        Game g = findGame(id, move);
        if (canSee(g.getId(), loginId)) {
            return g;
        }
        throw new WebApplicationException(Status.UNAUTHORIZED);
    }

    private boolean canSee(GameId id, String loginId) throws IOException {
        if (!id.isFinished()) {
            return true;
        }
        String logUser = quoridor.isLoggedIn(loginId);
        if (logUser == null) {
            return false;
        }
        if (logUser.equals(id.getWhite())) {
            return true;
        }
        if (logUser.equals(id.getBlack())) {
            return true;
        }
        User info = quoridor.getUsers().getUserInfo(loginId, logUser);
        if (info != null && info.hasPermission("games")) {
            return true;
        }
        return false;
    }

    @PUT
    @Path("{id}")
    @Produces({ MediaType.APPLICATION_JSON, MediaType.TEXT_XML })
    public GameId applyMove(
        @QueryParam("loginID") String loginId,
        @PathParam("id") String id,
        @QueryParam("player") String player,
        @QueryParam("move") String move,
        @QueryParam("comment") String comment
    ) throws IllegalPositionException {
        String logUser = quoridor.isLoggedIn(loginId);
        if (logUser == null) {
            throw new WebApplicationException(Status.UNAUTHORIZED);
        }
        if (!logUser.equals(player)) {
            throw new WebApplicationException(Status.UNAUTHORIZED);
        }
        if (comment == null && move == null) {
            throw new WebApplicationException(Status.BAD_REQUEST);
        }

        Game g = findGame(id);
        if (g == null) {
            throw new IllegalArgumentException("Unknown game " + id);
        }
        if (move != null) {
            Move m = Move.valueOf(move);
            g.apply(player, m, new Date());
        }
        if (comment != null) {
            g.comment(player, comment, new Date());
        }
        try {
            storeGame(g);
        } catch (IOException ex) {
            LOG.log(Level.WARNING, "Cannot store game " + id, ex);
        }
        return g.getId();
    }

    @GET
    @Produces({MediaType.APPLICATION_JSON, MediaType.TEXT_XML })
    public List<GameId> listGames(
        @DefaultValue("") @QueryParam("loginID") String loginId,
        @DefaultValue("") @QueryParam("status") String status
    ) throws IOException {
        List<GameId> arr = new ArrayList<GameId>(games.size());
        for (Game g : games) {
            if (!canSee(g.getId(), loginId)) {
                continue;
            }
            if (status.length() == 0 || g.getId().getStatus().toString().equals(status)) {
                arr.add(g.getId());
            }
        }
        Collections.sort(arr, GameId.NEWEST_FIRST);
        return arr;
    }

    public List<Game> getGames() {
        return games;
    }

    private Game findGame(String id) {
        for (Game g : games) {
            if (g.getId().getId().equals(id)) {
                return g;
            }
        }
        return null;
    }
    private Game findGame(String id, int move) {
        Game g = findGame(id);
        if (g == null) {
            throw new IllegalArgumentException("Unknown game " + id);
        }
        try {
            return move == -1 ? g : g.snapshot(move);
        } catch (IllegalPositionException ex) {
            Logger.getLogger(Games.class.getName()).log(Level.SEVERE, null, ex);
            return null;
        }
    }

    private static final Pattern saidWho = Pattern.compile("# *([^ ]*) *@(.*):$");

    private Game readGame(File f) throws IOException {
        InputStream is = new FileInputStream(f);
        BufferedReader r = new BufferedReader(new InputStreamReader(is, "UTF-8"));
        String white = null;
        String black = null;
        Game g = null;
        String who = null;
        String when = null;
        for (;;) {
            String line = r.readLine();
            if (line == null) {
                line = "finish";
            }
            line = line.trim();
            if (line.length() == 0) {
                continue;
            }
            if (line.startsWith("# white: ")) {
                white = line.substring(9);
                continue;
            }
            if (line.startsWith("# black: ")) {
                black = line.substring(9);
                continue;
            }
            if (line.startsWith("#")) {
                Matcher m =saidWho.matcher(line);
                if (m.matches()) {
                    who = m.group(1);
                    when = m.group(2);
                    continue;
                }
                if (g == null) {
                    continue;
                }
                if (line.startsWith("# ")) {
                    line = line.substring(2);
                } else {
                    line = line.substring(1);
                }
                Date d = new Date();
                try {
                    if (when != null) {
                        d = new Date(Date.parse(when));
                    }
                } catch (IllegalArgumentException ex) {
                    LOG.warning("Unparseable date " + when + " in " + f);
                }
                g.comment(who, line, d);
                who = null;
                when = null;
                continue;
            }
            if (white == null || black == null) {
                throw new IOException("Missing white and black identification in " + f);
            }
            if (g == null) {
                GameId id = new GameId(f.getName(), white, black, new Date(f.lastModified()), new Date(f.lastModified()), GameStatus.whiteMove, 0, false);
                g = new Game(id);
            }
            int hash = line.indexOf('#');
            if (hash >= 0) {
                line = line.substring(0, hash);
            }
            if (line.equals("finish")) {
                break;
            }
            String[] moves = line.split(" ");
            if (moves.length == 0) {
                continue;
            }
            if (moves.length > 2) {
                throw new IOException("Too much moves on line: " + line);
            }
            try {
                if (!"...".equals(moves[0])) {
                    g.apply(white, Move.valueOf(moves[0]), null);
                }
                if (moves.length == 2) {
                    g.apply(black, Move.valueOf(moves[1]), null);
                }
            } catch (IllegalPositionException ex) {
                throw new IOException("Wrong move: " + ex.getMessage());
            }
        }
        if (g == null) {
            throw new IOException("No moves in " + f);
        }
        return g;
    }

    private void storeGame(Game g) throws IOException {
        dir.mkdirs();
        File f = new File(dir, g.getId().getId());
        storeGame(g, f);
    }

    final void storeGame(Game g, File f) throws IOException {
        FileOutputStream os = new FileOutputStream(f);
        Writer w = new BufferedWriter(new OutputStreamWriter(os, "UTF-8"));
        PrintWriter pw = new PrintWriter(w);
        pw.println("# white: " + g.getId().getWhite());
        pw.println("# black: " + g.getId().getBlack());
        pw.println("# status: " + g.getId().getStatus());
        int cnt = 0;
        boolean separate = true;
        for (CommentedMove m : g.getMoves()) {
            if (!separate && cnt % 2 == 1) {
                pw.print("... ");
            }
            separate = true;
            pw.print(m.getMove().toString());
            List<Note> notes = m.getComments();
            if (notes != null) {
                separate = false;
                pw.println();
                for (Note n : notes) {
                    pw.print ("# ");
                    pw.print(n.getWho());
                    pw.print("@");
                    pw.print(n.getWhen());
                    pw.println(":");
                    for (String l : n.getComment().split("\n")) {
                        pw.print("# ");
                        pw.println(l);
                    }
                }
            }

            cnt++;

            if (separate) {
                if (cnt % 2 == 0) {
                    pw.println();
                } else {
                    pw.print(' ');
                }
            }
        }
        pw.println();
        pw.println();
        pw.flush();
        pw.close();
        w.close();
    }
}
